/*
 * Created on Jun 8, 2005
 * @author Rafael Santos (rafael.santos@lac.inpe.br)
 * 
 * Part of the Java Advanced Imaging Stuff site
 * (http://www.lac.inpe.br/~rafael.santos/Java/JAI)
 * 
 * STATUS: Complete.
 * 
 * Redistribution and usage conditions must be done under the
 * Creative Commons license:
 * English: http://creativecommons.org/licenses/by-nc-sa/2.0/br/deed.en
 * Portuguese: http://creativecommons.org/licenses/by-nc-sa/2.0/br/deed.pt
 * More information on design and applications are on the projects' page
 * (http://www.lac.inpe.br/~rafael.santos/Java/JAI).
 */

import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.awt.image.renderable.ParameterBlock;
import java.util.Arrays;
import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.media.jai.RasterFactory;
import javax.media.jai.RenderedOp;
import javax.media.jai.TiledImage;
import javax.media.jai.iterator.RandomIter;
import javax.media.jai.iterator.RandomIterFactory;

/**
 * This class implements a basic K-Means clustering algorithm as an image
 * processing task. This implementation tries to speed things up, but needs
 * to keep all image data on memory.
 * This implementation deals only with integer-like pixel data but can
 * cluster N-dimensional data.
 * One thing to remember about K-Means: it is possible that some clusters
 * will become "empty". This implementation does not deal with any solution
 * to this case, the clusters will simply become empty and will not be
 * further processed.
 * IMPORTANT: This implementation does not lock access to the data while the
 * clustering is being done.
 */
public class KMeansImageClustering extends ImageProcessingTask
  {
  // The output (clustered) image.
  private TiledImage pOutput;
  // The input image dimensions.
  private int width,height,numBands;
  // Some clustering parameters.
  private int maxIterations,numClusters;
  // The iteration counter will be global so we can get its value on the
  // middle of the clustering process.
  private int iteration;
  // A metric of clustering "quality".
  private double sumOfDistances;
  // A small value, if the difference of the cluster "quality" does not
  // changes beyond this value, we consider the clustering converged.
  private double epsilon;
  // Flags and counters.
  private boolean finished = false;
  private long position;
  // The cluster centers.
  private float[][] clusterCenters;

  // The centers in pixel space.
  private float[] clusterCentersX;
  private float[] clusterCentersY;

  // The cluster assignment counter.
  private int[] clusterAssignmentCount;
  // A big array with all the input data and a small one for a pixel.
  private int[] inputData,aPixel;
  // A big array with the output data (cluster indexes).
  private short[][] outputData;
  // keep some info 
  private KMeansClusterDescriptor[] descriptors; 

  private double global_com;
  private double global_col; 
  private double global_DIS;
  private double global_COL;

  /**
   * The constructor for the class, which sets the input image, the number of
   * desired clusters, the maximum number of iterations, a value that will be
   * used to decide whether the convergence has stopped and the initial
   * clusters centers initialization mode. It also allocates needed memory.
   * @param pInput the input planar image.
   * @param numClusters the desired number of clusters.
   * @param maxIterations the maximum number of iterations.
   * @param epsilon a small value used to verify if clustering has converged.
   * @param initMode the initialization mode: 'R' for random values, 'S' for
   *        evenly spaced values, 'D' for randomly sampled data values.
   */
  public KMeansImageClustering(PlanarImage pInput,int numClusters,int maxIterations,
                          double epsilon,char initMode, float[][]iniCenters)
    {
    // Get the image dimensions.
    width = pInput.getWidth();
    height = pInput.getHeight();
    numBands = pInput.getSampleModel().getNumBands();
    // Create the output image based on the input image's dimensions.
    pOutput = new TiledImage(pInput.getMinX(),pInput.getMinY(),width,height,
                             pInput.getMinX(),pInput.getMinY(),pInput.getSampleModel(),pInput.getColorModel());
    // Get some clustering parameters.
    this.numClusters = numClusters;
    this.maxIterations = maxIterations;
    this.epsilon = epsilon;
    iteration = 0;
    // We need arrays to store the clusters' centers and assignment counts.
    clusterCenters = new float[numClusters][numBands];
    clusterAssignmentCount = new int[numClusters];

     clusterCentersX = new float[numClusters];
     clusterCentersY = new float[numClusters];

    // Gets the raster for the input image.
    Raster raster = pInput.getData();
    // Gets the whole image data on memory. Get memory for a single pixel too.
    inputData = new int[width*height*numBands];
    aPixel = new int[numBands];
    // Gets memory for the output data (cluster indexes).
    outputData = new short[width][height];
    raster.getPixels(0,0,width,height,inputData);
    // Initialize the clusters centers, depending on the initialization mode.
    // Evenly spaced values (over main diagonal).
    if ((initMode == 's') || (initMode == 'S'))
      {
      // Discover the extrema value for the image data.
      double[] extrema = getExtrema(pInput);
      double delta = (extrema[1]-extrema[0])/(numClusters-1);
      for(int cluster=0;cluster<numClusters;cluster++)
        {
        for(int band=0;band<numBands;band++)
          clusterCenters[cluster][band] = (float)extrema[0];
        extrema[0] += delta;
        }
      }


     // from external
     if ((initMode == 'T') || (initMode == 't'))
      {

      for(int cluster=0;cluster<numClusters;cluster++)
        {
        for(int band=0;band<numBands;band++)
          clusterCenters[cluster][band] = iniCenters[cluster][band];
        }
     }





    // Randomly sampled data values.
    else if ((initMode == 'd') || (initMode == 'D'))
      {
      // Create a Iterator to get the samples.
      RandomIter iterator = RandomIterFactory.create(pInput,null);
      for(int cluster=0;cluster<numClusters;cluster++)
        {
        int rx = (int)(Math.random()*width);
        int ry = (int)(Math.random()*height);
        iterator.getPixel(rx,ry,aPixel);
        for(int band=0;band<numBands;band++)
          clusterCenters[cluster][band] = aPixel[band];
        }  
      }
    // Random values, using the maximum of the data values as limit.
    else // if ((initMode == 'r') || (initMode == 'R'))
      {
      // Discover the maximum value for the image data.
      double[] extrema = getExtrema(pInput);
      for(int cluster=0;cluster<numClusters;cluster++)
        for(int band=0;band<numBands;band++)
           clusterCenters[cluster][band] = (float)(extrema[1]*Math.random());
      }
    }

 /**
  * This method returns a two-element array with the minimum and maximum
  * values found in a PlanarImage.
  * @param image the input PlanarImage.
  * @return an array with two positions: the first is the minima of the image
  *         and the second is the maxima.
  */
  private double[] getExtrema(PlanarImage image)
    {
    ParameterBlock pbMaxMin = new ParameterBlock();
    pbMaxMin.addSource(image);
    RenderedOp extremaOp = JAI.create("extrema", pbMaxMin);
    // Must get the extrema of all bands !
    double[] allMins = (double[])extremaOp.getProperty("minimum");
    double[] allMaxs = (double[])extremaOp.getProperty("maximum");
    double minValue = allMins[0];
    double maxValue = allMaxs[0];
    for(int v=1;v<allMins.length;v++)
      {
      if (allMins[v] < minValue) minValue = allMins[v];
      if (allMaxs[v] > maxValue) maxValue = allMaxs[v];
      }
    return new double[] {minValue,maxValue};
    }
 
 /**
  * This method performs the bulk of the processing. It runs the classic
  * K-Means clustering algorithm:
  * 1 - Scan the image and calculate the assignment vector.
  * 2 - Scan the assignment vector and recalculate the cluster centers.
  * 3 - Calculate statistics and repeat from 1 if needed.
  */
  public void run()
    {
    double lastSumOfDistances=0;
    iterations: // Label for main loop.
      for(iteration=0;iteration<maxIterations;iteration++)
        {
        // 0 - Clean the cluster assignment vector.
        Arrays.fill(clusterAssignmentCount,0);
        // 1 - Scan the image and calculate the assignment vector.
        for(int h=0;h<height;h++)
          for(int w=0;w<width;w++)
            {
            // Where does the individual pixel data start ?
            int index = (h*width+w)*numBands;
            // Gets the class for this pixel.
            for(int band=0;band<numBands;band++) aPixel[band] = inputData[index+band];
            short aClass = getClassFor(aPixel);
            // Update the position index.
            position += numBands + numClusters*numBands;
            outputData[w][h] = aClass;
            clusterAssignmentCount[aClass]++;
            }
        // 2 - Scan the assignment vector and recalculate the cluster centers.
        for(int cluster=0;cluster<numClusters;cluster++) Arrays.fill(clusterCenters[cluster],0f);
        // Update the position index.
        position += numClusters;
        for(int h=0;h<height;h++)
          for(int w=0;w<width;w++)
            {
            int theCluster = outputData[w][h];
            for(int band=0;band<numBands;band++)
              {
              int index = (h*width+w)*numBands;
              clusterCenters[theCluster][band] += inputData[index+band];
              }
            // Update the position index.
            position += numBands;
            }
        // 2a - Recalculate the centers.
        for(int cluster=0;cluster<numClusters;cluster++)
          if (clusterAssignmentCount[cluster] > 0)
            for(int band=0;band<numBands;band++)
              clusterCenters[cluster][band]/=clusterAssignmentCount[cluster];
        // Update the position index.
        position += numClusters*numBands;
        // 3 - Calculate statistics and repeat from 1 if needed.
        sumOfDistances = 0;
        for(int h=0;h<height;h++)
          for(int w=0;w<width;w++)
            {
            // To which class does this pixel belong ?
            short pixelsClass = outputData[w][h];
            // Calculate the distance between this pixel's values and its
            // assigned cluster center values.
            double distance = 0;
            int index = (h*width+w)*numBands;
            for(int band=0;band<numBands;band++)
              {
              double e1 = inputData[index+band];
              double e2 = clusterCenters[pixelsClass][band];
              double diff = (e1-e2)*(e1-e2);
              distance += diff;
              }
            distance = Math.sqrt(distance);
            sumOfDistances += distance;
            // Update the position index.
            position += numBands;
            }
        // Is it converging ?
        if (iteration > 0)
          if (Math.abs(lastSumOfDistances-sumOfDistances) < epsilon) break iterations;


        lastSumOfDistances = sumOfDistances;
        } // end of the iterations loop.
    finished = true;
    // Means that all calculations are done, too.
    position = getSize();


     for(int h=0;h<height;h++)
      for(int w=0;w<width;w++)
        {
        // Get the class (cluster center) for that pixel.
        short aClass = outputData[w][h];
        clusterCentersX[(int)aClass] = clusterCentersX[(int)aClass]+w;
        clusterCentersY[(int)aClass] = clusterCentersY[(int)aClass]+h;

  //       System.out.println("Cluster h="+Float.toString(h)+
  //                        " w="+Float.toString(w)+ " ID="+Short.toString(aClass));

        }
   
     for(int d=0;d<numClusters;d++)
      {
      clusterCentersX[d] = clusterCentersX[d]/clusterAssignmentCount[d];
      clusterCentersY[d] = clusterCentersY[d]/clusterAssignmentCount[d];
      }
 

    // get some info
    descriptors = new KMeansClusterDescriptor[numClusters];
    for(int d=0;d<numClusters;d++)
      {
      descriptors[d] = new KMeansClusterDescriptor();
      descriptors[d].setCentroid(clusterCenters[d]);
      descriptors[d].setCentroidXY( clusterCentersX[d], clusterCentersY[d] );

   //  System.out.println("Average h="+Float.toString(clusterCentersX[d])+
   //                       " w="+Float.toString(clusterCentersY[d]));

      descriptors[d].setNumberOfAssignedPoints(clusterAssignmentCount[d]);
      }


      int ilargest = KMeansRegionClusterer.getLargestClusterIndex(descriptors);
 
      global_com=0;
      // get compactness in terms of normalized ratios 
      // remove largest cluster (i.e. background)
      for(int d=0;d<numClusters;d++)
      {
      if (d  == ilargest) continue;
 
      double cx=clusterCentersX[d] /(double)width;
      double cy=clusterCentersY[d] /(double)height;
      double xcom=0; 
      double ccom=0;

      for(int h=0;h<height;h++)
      for(int w=0;w<width;w++)
        {
        short aClass = outputData[w][h];
        if ( (int)aClass == d) {
        double e1x = (double)w /width;
        double e1y = (double)h /height;
        double diff = (e1x-cx)*(e1x-cx) + (e1y-cy)*(e1y-cy);
        xcom=xcom+Math.sqrt(diff); 


         // this is for color compactness
         int index = (h*width+w)*numBands;
         // Gets the class for this pixel.
         for(int band=0;band<numBands;band++) aPixel[band] = inputData[index+band];
                diff = (clusterCenters[d][0]-aPixel[0])*(clusterCenters[d][0]-aPixel[0]) +
                       (clusterCenters[d][1]-aPixel[1])*(clusterCenters[d][1]-aPixel[1]) +
                       (clusterCenters[d][2]-aPixel[2])*(clusterCenters[d][2]-aPixel[2]);

          ccom=ccom+Math.sqrt(diff);
        }
       }
       xcom=xcom/clusterAssignmentCount[d];    
       global_com=global_com+xcom;
       descriptors[d].setCompactness(xcom); 

       ccom=ccom/clusterAssignmentCount[d];
       descriptors[d].setCompactnessCol(ccom);
       global_col=global_col+ccom;

      } 
      global_com=global_com/(numClusters-1); 
      global_col=global_col/(numClusters-1);





      double dd=0;
      for(int d1=0;d1<numClusters-1;d1++) { 
        double cx1=clusterCentersX[d1] /(double)width;
        double cy1=clusterCentersY[d1] /(double)height;
 

      for(int d2=d1+1;d2<numClusters;d2++) { 
          double cx2=clusterCentersX[d2] /(double)width;
          double cy2=clusterCentersY[d2] /(double)height;
          double dis=(cx1-cx2)*(cx1-cx2)+(cy1-cy2)*(cy1-cy2);
          dd=dd+Math.sqrt(dis);  
       }
      }
      global_DIS=2*dd/((numClusters-1)*numClusters);


      dd=0;
      for(int d1=0;d1<numClusters-1;d1++) {
        double cx1=clusterCenters[d1][0];
        double cy1=clusterCenters[d1][1];
        double cz1=clusterCenters[d1][2];


      for(int d2=d1+1;d2<numClusters;d2++) {
        double cx2=clusterCenters[d2][0];
        double cy2=clusterCenters[d2][1];
        double cz2=clusterCenters[d2][2];
        double dis=(cx1-cx2)*(cx1-cx2)+(cy1-cy2)*(cy1-cy2)+(cz1-cz2)*(cz1-cz2);
        dd=dd+Math.sqrt(dis); 
       }
      }
      global_COL=2*dd/((numClusters-1)*numClusters);

    }



  /**
  * get description 
  * @return the class (cluster index) for a pixel.
  */
   public KMeansClusterDescriptor[] getDescriptor() {
    return descriptors;
    }



  /**
  * get guality depending (sum over compactness) 
  * @return guality 
  */
   public double getQualityXY() {
    return global_com;
    }

/**
  * get guality depending in color space 
  * @return guality
  */
   public double getQualityCol() {
    return global_col;
    }

/**
  * get avearage color centers
  * @return guality
  */
   public double getDistanceCol() {
    return global_COL;
    }

/**
  * get avearage XY centers
  * @return guality
  */
   public double getDistanceXY() {
    return global_DIS;
    }


 /**
  * This auxiliary method gets the class for a pixel vector.
  * @return the class (cluster index) for a pixel.
  */
  private short getClassFor(int[] pixel)
    {
    // Let's compare this pixel data with all the cluster centers.
    float closestSoFar = Float.MAX_VALUE;
    short classSoFar = 0;
    for(short cluster=0;cluster<numClusters;cluster++)
      {
      // Calculate the (quick) distance from this pixel to that cluster center.
      float distance = 0f;
      for(int band=0;band<numBands;band++)
        distance += Math.abs(clusterCenters[cluster][band]-pixel[band]);
      if (distance < closestSoFar)
        {
        closestSoFar = distance;
        classSoFar = cluster;
        }
      }
    return classSoFar;
    }

 /**
  * This method returns the output image. It can return an image which is
  * being processed while this method run, so results are not guaranteed
  * to be final (and can be very strange if the image is large).
  * @return the (possibly temporary) clustering results.
  */
  public TiledImage getOutput()
    {
    // We must dump the crossed contents of the cluster centers and the
    // cluster indexes into the resulting image.
    // Create a SampleModel for the output data.
    SampleModel sampleModel =
      RasterFactory.createBandedSampleModel(DataBuffer.TYPE_INT,width,height,numBands);
    // Create a WritableRaster using that sample model.
    WritableRaster raster = RasterFactory.createWritableRaster(sampleModel,new Point(0,0));
    // A pixel array will contain all bands for a specific x,y.
    int[] pixelArray = new int[numBands];
    // For all pixels in the image...
    for(int h=0;h<height;h++)
      for(int w=0;w<width;w++)
        {
        // Get the class (cluster center) for that pixel.
        short aClass = outputData[w][h];
        // Fill the array with the cluster center.
        for(int band=0;band<numBands;band++) pixelArray[band] = (int)clusterCenters[aClass][band];
        // Put it on the raster.
        raster.setPixel(w,h,pixelArray);
        }
    // Set the raster on the output image.
    pOutput.setData(raster);
    return pOutput;
    }

 /**
  * This method returns the estimated size (steps) for this task.
  * The value is, of course, an approximation, just so we will be able to
  * give the user a feedback on the processing time. In this case, the
  * value is calculated as the number of loops in the run() method.
  */
  public long getSize()
    {
    // Return the estimated size for this task:
    return (long)maxIterations* // The maximum number of iterations times
      (
       (width*height*(numBands+numClusters*numBands))+ // Step 1 of run()
       (numClusters+width*height*numBands)+ // Step 2 of method run()
       (numClusters*numBands)+  // Step 2a of method run()
       (width*height*numBands) // Step 3 of the method run()
      );
    }

 /**
  * This method returns a measure of the progress of the algorithm.
  */
  public long getPosition()
    {
    return position;
    }

 /**
  * This method returns true if the clustering has finished.
  */
  public boolean isFinished()
    {
    return finished;
    }


  /**
  * Returns the centroid of this cluster as a array of doubles.
  * @return the cluster's centroid.
  */
  public float[][] getClusterCenters()
    {
    return clusterCenters;
    }


  /**
  * Print cluster index 
  */
  public void  printClusterIndex() { 

   for(int h=0;h<height;h++)
      for(int w=0;w<width;w++)
        {
        // Get the class (cluster center) for that pixel.
        short aClass = outputData[w][h];
        System.out.println("x="+Integer.toString(h)+" "+
           Integer.toString(w) +" " +Short.toString(aClass));
 
        }
    }

 /**
  * This method returns some information about the task (its progress).
  * @return some information about the task
  */
  public String getInfo()
    {
    if (!finished) return "Iteration "+(iteration+1)+" of "+maxIterations;
    else return "Clustering converged.";
    }



/*	  *//**
			 * This method returns the Compactness
			 * 
			 * @param indat[][]
			 *            data matrix
			 * @param numClusters
			 *            number of clusters
			 * @param clusterCenters[][]
			 *            cluster centers
			 * @return compactness
			 */

	public double Compactness(double indat[][], int[] assignment,
			int numClusters, double clusterCenters[][], int NRow, int Dim) {

		int nrow = NRow;
		int ncol = Dim;
		double[] aPixel = new double[ncol];

		double[] dsize = new double[numClusters];
		double[] isize = new double[numClusters];
		for (int j = 0; j < numClusters; j++) {
			dsize[j] = 0;
			isize[j] = 0;
		}

		double cs = 0;
		double maxSize = 0;
		// For all data values and clusters
		for (int h = 0; h < nrow; h++) {
			int current = assignment[h];

			if (current >= 0) {
				// Get the current pixel data.
				for (int b = 0; b < ncol; b++) aPixel[b] = indat[h][b];
					double ddd=calcDistance(aPixel, clusterCenters[current]);					
					dsize[current] = dsize[current]+ddd;
					maxSize= Math.max(maxSize, ddd );
					isize[current]++;
			}
		}

		
		// average size
		double csumm=0;
		for (int j = 0; j < numClusters; j++) {
			csumm = csumm + dsize[j] / (double) isize[j];
		}
		csumm=csumm/numClusters;
		


		// Calculate minimum distance between ALL clusters
		double minDist = Double.MAX_VALUE;
		for (int c1 = 0; c1 < numClusters - 1; c1++)
			for (int c2 = c1 + 1; c2 < numClusters; c2++) {
				double distance = calcDistance(clusterCenters[c1],
						clusterCenters[c2]);
				minDist = Math.min(minDist, distance);
			}

		cs = maxSize / minDist;

 // System.out.println(" maxSize="+maxSize+ " min DIS="+minDist
//		 +" average size="+csumm);
 
		return cs;
	}



/**
	 * This method calculates the squared distance between two N-dimesional
	 * vectors.
	 * 
	 * @param a1
	 *            the first data vector.
	 * @param a2
	 *            the second data vector.
	 * @return the squared distance between those vectors.
	 */
	public double calcSquaredDistance(double[] a1, double[] a2) {
		double distance = 0f;
		for (int e = 0; e < a1.length; e++)
			distance += (a1[e] - a2[e]) * (a1[e] - a2[e]);
		return (double) distance;
	}

	/**
	 * This method calculates the Euclidean distance between two N-ncolsional
	 * vectors.
	 * 
	 * @param a1
	 *            the first data vector.
	 * @param a2
	 *            the second data vector.
	 * @return the Euclidean distance between those vectors.
	 */
	public double calcDistance(double[] a1, double[] a2) {
		double distance = 0f;
		for (int e = 0; e < a1.length; e++)
			distance += (a1[e] - a2[e]) * (a1[e] - a2[e]);
		return (double) Math.sqrt(distance);
	}












  }
