
/**
 * static class StatTools
 * 
 * Provides statistics functions
 *
 * Recipes taken from:
 * S. Brandt, Datenanalyse, 3. Auflage,
 * BI Wissenschaftsverlag, Mannheim/Leipzig/Wien/Zuerich, 1992 
 *
 * @author W.Walkowiak, 08/08/00
 * @version $Id: StatTools.java,v 1.4 2002/06/26 17:49:16 tonyj Exp $
 *
 */
public final class StatTools
{
    /**
     * Integrated Chi2 probability 
     * with ndf free parameters.
     *
     * source: S.Brandt, Datenanalyse, p.549
     */
    public static double chi2Df(double chi2, int ndf)
    {
	double a = 0.5*(double)ndf;
	return incGammaFunc(a,0.5*chi2);
    }






 // some  staff 
     public static double prob(double chi2, int ndf) {  
           double x= Math.sqrt(chi2)/(ndf);
           x=1.0/Math.exp(x);
           return x;
     }





    /**
     * Incomplete gamma function
     *
     * source: S.Brandt, Datenanalyse, p.564f
     */
    public static double incGammaFunc(double a, double x)
    {
	final double big = 500.;
	final double epsiln = 1.E-6;

	double value, help;

	double aloggm=lnGamma(a);  
	if ( x <= a+1. )
	{
	    // series development
	    double f = 1./a;
	    double s = f;
	    double anum = a;
	loop1:
	    for ( int i=1; i<=100; i++ )
	    {
		anum = anum+1.;
		f = x*f/anum;
		s = s+f;
		if ( f < epsiln ) break loop1;
	    }
	    if ( x < epsiln ) 
		value = 0.;
	    else
	    {
		help = a*Math.log(x)-x-aloggm;
		if ( Math.abs(help) >= big )
		    value = 0.;
		else
		    value = s*Math.exp(help);
	    }
	}
	else
	{
	    // continued fraction
	    double a0    = 0.;
	    double b0    = 1.;
	    double a1    = 1.;
	    double b1    = x;
	    double cf    = 0.;
	    double fnorm = 1.;
	    double a2j, a2j1, b2j, b2j1; 
	    double cfnew = 0.;
	loop2:
	    for ( int j=1; j<=100; j++ ) 
	    {
		a2j  = (double)j-a;
		a2j1 = (double)j;
		b2j  = 1.;
		b2j1 = x;
		a0   = (b2j*a1+a2j*a0)*fnorm;
		b0   = (b2j*b1+a2j*b0)*fnorm;
		a1   = b2j1*a0+a2j1*a1*fnorm;
		b1   = b2j1*b0+a2j1*b1*fnorm;
		if ( b1 != 0. ) 
		{
		    fnorm = 1./b1;
		    cfnew = a1*fnorm;
		    if ( Math.abs(cf-cfnew)/cf < epsiln ) break loop2;
		    cf = cfnew;
		}
	    }
	    help = a*Math.log(x)-x-aloggm;
	    if ( Math.abs(help) >= big ) 
		value = 1.;
	    else
		value = 1.-Math.exp(help)*cfnew;
	}
	return value;
    }
    
    /**
     * logarithmic Gamma function ln Gamma(x)
     *
     * source: S.Brandt, Datenanalyse, p.559f
     */
    public static double lnGamma(double x)
    {
	double[] c = {76.18009173,-86.50532033,
		      24.01409822,-1.231739516,
		      0.120858003E-2,-0.536382E-5};
	boolean reflec;
	final double RTWOPI=2.506628275;
	
	double xx, value, g;

	if ( x >= 1. ) 
	{
	    reflec = false;
	    xx = x-1.;
	}
	else
	{
	    reflec = true;
	    xx = 1.-x;
	}
	double xh   = xx+0.5;
	double xgh  = xx+5.5;
	double s    = 1.;
	double anum = xx;
	for ( int i=0; i<6; i++) 
	{
	    anum = anum+1.;
	    s    = s+c[i]/anum;
	}
	s = s*RTWOPI;
	g = xh*Math.log(xgh)+Math.log(s)-xgh;
	if ( reflec ) 
	    value = Math.log(Math.PI*xx)-g-Math.log(Math.sin(Math.PI*xx));
	else
	    value = g;
	return value;
    }

    /**
     * Chi2 probability density function f(Chi2)
     *
     * source: S.Brandt, Datenanalyse, p.548f
     */
    public static double chi2Pdf(double chi2, int n)
    {
	double value =0.;
	if ( chi2 != 0. || n > 2 ) 
	{
	    double b = gammaFunc(0.5*(double)n)*Math.sqrt(Math.pow(2.,(double)n));
	    double c = Math.sqrt(Math.pow(chi2,(double)(n-2)));
	    double e = Math.exp(-0.5*chi2);
	    value    = e*c/b;
	}
	else
	{
	    if ( n == 2 ) value = 0.5;
	    if ( n == 1 ) value = 1.E10;
        }
	return value;
    }

    /**
     * Euler's Gamma function Gamma(x)
     *
     * source: S.Brandt, Datenanalyse, p.558f
     */
    static double gammaFunc(double x)
    {
	double[] c = {76.18009173,-86.50532033,
		      24.01409822,-1.231739516,
		      0.120858003E-2,-0.536382E-5};
	boolean reflec;
	final double RTWOPI=2.506628275;
	
	double xx, value, g;

	if ( x >= 1. ) 
	{
	    reflec = false;
	    xx = x-1.;
	}
	else
	{
	    reflec = true;
	    xx = 1.-x;
	}
	double xh   = xx+0.5;
	double xgh  = xx+5.5;
	double s    = 1.;
	double anum = xx;
	for ( int i=0; i<6; i++) 
	{
	    anum = anum+1.;
	    s    = s+c[i]/anum;
	}
	s = s*RTWOPI;
	g = Math.pow(xgh,xh)*s/Math.exp(xgh);
	if ( reflec ) 
	    value = Math.PI*xx/(g*Math.sin(Math.PI*xx));
	else
	    value = g;
	return value;
    }

}




