/*
 * Created on Dec 26, 2005
 * @author Rafael Santos (rafael.santos@lac.inpe.br)
 * 
 * Part of the Java Advanced Imaging Stuff site
 * (http://www.lac.inpe.br/~rafael.santos/Java/JAI)
 * 
 * STATUS: Complete.
 * 
 * Redistribution and usage conditions must be done under the
 * Creative Commons license:
 * English: http://creativecommons.org/licenses/by-nc-sa/2.0/br/deed.en
 * Portuguese: http://creativecommons.org/licenses/by-nc-sa/2.0/br/deed.pt
 * More information on design and applications are on the projects' page
 * (http://www.lac.inpe.br/~rafael.santos/Java/JAI).
 */
package operators;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.RenderedImage;
import java.awt.image.WritableRaster;
import java.awt.image.renderable.ParameterBlock;
import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.media.jai.RenderedOp;
import javax.media.jai.iterator.RandomIter;
import javax.media.jai.iterator.RandomIterFactory;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import com.sun.media.jai.widget.DisplayJAI;

/**
 * This class shows how we can ensure that image subtraction results are not
 * clamped -- basically we convert the image data to integer before
 * subtraction (the image data is originally byte, therefore integers will
 * be more than enough to represent the results without clamping).
 * This class creates some patterns and allow the user to subtract them 
 * as byte images (with clamping).
 * The classes' user interface shows the input and output images. When the
 * user clicks on any pixel of the images, the subtraction result is presented
 * on a JLabel.
 */
public class Subtract extends JFrame implements MouseListener,ActionListener
  {
  // Constants for image size.
  private static final int width = 256;
  private static final int height = 256;
  // Input and output RenderedImages. 
  private RenderedImage input1,input2,output;
  // Components for displaying those images. 
  private DisplayJAI displayInput1,displayInput2,displayOutput;
  // Iterators to get those images pixels' data.
  private RandomIter iterInput1,iterInput2,iterOutput;
  private int[] pi1,pi2,pout;
  // A JLabel to show the results.
  private JLabel results;
  // A JCheckBox to set the mode (int or byte subtraction).
  private JCheckBox mode;
  
 /**
  * The constructor for the class, which will set its user interface
  * and initialize some data structures.
  */
  public Subtract()
    {
    super("Subtraction results");
    // Create a UI to see the results. We won't use DisplayThreeSynchronizedImages
    // since we want to add some events to do some inspection on the images data.
    mode = new JCheckBox("Use signed integer for subtraction",false);
    mode.addActionListener(this);
    getContentPane().add(mode,BorderLayout.NORTH);
    // A JPanel to hold the input and output images.
    JPanel panel = new JPanel(new FlowLayout());
    input1 = createFirstPattern(true);
    displayInput1 = new DisplayJAI(input1);
    displayInput1.addMouseListener(this);
    input2 = createSecondPattern(true);
    displayInput2 = new DisplayJAI(input2);
    displayInput2.addMouseListener(this);
    // We need something to display, let's create the subtracted image.
    ParameterBlock pb = new ParameterBlock();
    pb.addSource(input1); pb.addSource(input2);
    output = JAI.create("subtract", pb);
    displayOutput = new DisplayJAI(output); 
    displayOutput.addMouseListener(this);
    panel.add(displayInput1);
    panel.add(new JLabel("-"));
    panel.add(displayInput2);
    panel.add(new JLabel("="));
    panel.add(displayOutput);
    getContentPane().add(panel,BorderLayout.CENTER);
    // A JLabel to hold the results.
    results = new JLabel(" ");
    getContentPane().add(results,BorderLayout.SOUTH);
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    pack();
    setVisible(true);
    // Create the iterators. 
    iterInput1 = RandomIterFactory.create(input1, null);
    iterInput2 = RandomIterFactory.create(input2, null);
    iterOutput = RandomIterFactory.create(output, null);
    // Create some arrays to use with the iterators.
    pi1 = new int[input1.getSampleModel().getNumBands()];
    pi2 = new int[input2.getSampleModel().getNumBands()];
    pout = new int[output.getSampleModel().getNumBands()];
    }

 /**
  * This method will be called when the checkbox status is changed.
  */
  public void actionPerformed(ActionEvent e)
    {
    if (mode.isSelected()) 
      {
      // Create integer images.
      input1 = createFirstPattern(false);
      input2 = createSecondPattern(false);
      // Subtract with integer images
      ParameterBlock pb = new ParameterBlock();  
      pb.addSource(input1);
      pb.addSource(input2);
      output = JAI.create("subtract", pb);
      // This new output image may have negative values, let's create
      // a surrogate image for displaying.
      displayOutput.set(createSurrogate(output));
      }
    else // the checkbox is not selected, we do the subtraction without  
      {  // conversions and without creating surrogate images
      // Create byte images.
      input1 = createFirstPattern(true);
      input2 = createSecondPattern(true);
      // Subtract with integer images
      ParameterBlock pb = new ParameterBlock();  
      pb.addSource(input1);
      pb.addSource(input2);
      output = JAI.create("subtract", pb);
      // Set the output display image.
      displayOutput.set(output);
      }
    // Reassign the iterators. The input are not really necessary.
    iterInput1 = RandomIterFactory.create(input1, null);
    iterInput2 = RandomIterFactory.create(input2, null);
    iterOutput = RandomIterFactory.create(output, null);
    }
 
 /**
  * This method will be called when we click with the mouse
  * button on one of the images.
  */
  public void mouseClicked(MouseEvent e)
    {
    int x = e.getX(); int y = e.getY();
    iterInput1.getPixel(x,y,pi1);
    iterInput2.getPixel(x,y,pi2);
    iterOutput.getPixel(x,y,pout);
    results.setText(pi1[0]+" - "+pi2[0]+" = "+pout[0]);
    }

  // Must include these methods to satisfy MouseListener.
  public void mousePressed(MouseEvent e) { } // nop
  public void mouseReleased(MouseEvent e) { } // nop
  public void mouseEntered(MouseEvent e) { } // nop
  public void mouseExited(MouseEvent e) { } // nop
  
  // This method creates a surrogate image -- one which is
  // used to display data which originally was outside the 
  // [0-255] range.
  private RenderedImage createSurrogate(RenderedImage image)
    {
    ParameterBlock pb = new ParameterBlock();  
    pb.addSource(image);  
    RenderedOp extrema = JAI.create("extrema", pb);  
    // Must get the extrema of all bands.  
    double[] allMins = (double[])extrema.getProperty("minimum");  
    double[] allMaxs = (double[])extrema.getProperty("maximum");  
    double minValue = allMins[0];  
    double maxValue = allMaxs[0];  
    for(int v=1;v<allMins.length;v++)  
      {  
      if (allMins[v] < minValue) minValue = allMins[v];  
      if (allMaxs[v] > maxValue) maxValue = allMaxs[v];  
      }  
    // Rescale the image with the parameters
    double[] multiplyBy = new double[1]; multiplyBy[0] = 255./(maxValue-minValue);  
    double[] addThis    = new double[1]; addThis[0]    = -minValue*multiplyBy[0];  
    // Now we can rescale the pixels gray levels:
    pb = new ParameterBlock();  
    pb.addSource(image);  
    pb.add(multiplyBy);  
    pb.add(addThis);  
    RenderedImage surrogateImage = JAI.create("rescale", pb); 
    // Let's convert the data type for displaying.
    pb = new ParameterBlock();  
    pb.addSource(surrogateImage);  
    pb.add(DataBuffer.TYPE_BYTE);  
    surrogateImage = JAI.create("format", pb);  
    return surrogateImage;
    }
  
  // This method creates a RenderedImage with a degradé pattern.
  private static RenderedImage createFirstPattern(boolean isByte)
    {
    // Create the image.
    BufferedImage image = new BufferedImage(width,height,BufferedImage.TYPE_BYTE_GRAY);
    // We need its raster to set the pixels' values.
    WritableRaster raster = image.getRaster();
    // Put the pixels on the raster. 
    for(int h=0;h<height;h++)
      for(int w=0;w<width;w++)
        raster.setSample(w,h,0,(h*w)/width);
    // Return with conversion, if needed.
    if (isByte) return image;
    else return convertToInt(image);
    }
  
  // This method creates a RenderedImage with a checkboard pattern.
  private static RenderedImage createSecondPattern(boolean isByte)
    {
    // Create a Buffered Image.
    BufferedImage image = new BufferedImage(width,height,BufferedImage.TYPE_BYTE_GRAY);
    // We need its raster to set the pixels' values.
    WritableRaster raster = image.getRaster();
    // Put the pixels on the raster. 
    for(int h=0;h<height;h++)
      for(int w=0;w<width;w++)
        if (((h/64) + (w/64)) % 2 == 0) raster.setSample(w,h,0,0);
        else raster.setSample(w,h,0,255);
    // Return with conversion, if needed.
    if (isByte) return image;
    else return convertToInt(image);
    }
  
  // This method formats a RenderedImage with integer data type.
  private static RenderedImage convertToInt(BufferedImage image)
    {
    ParameterBlock pb = new ParameterBlock();  
    pb.addSource(image);  
    pb.add(DataBuffer.TYPE_INT);  
    PlanarImage newImage = JAI.create("format", pb);  
    return newImage;
    }
  
  /**
   * @param args the command-line arguments, which will be ignored.
   */
  public static void main(String[] args) 
    {
    // Create and launches the UI.
    new Subtract();
    }

  }